/**
 * <copyright>
 *
 * Copyright (c) 2002 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * plugins/org.eclipse.xsd/src/org/eclipse/xsd/util/XSDParser.java, xsd, org.eclipse.111, 20031120_1149WL
 * @version 1.9 11/20/03
 */
package org.eclipse.xsd.util;


import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;

import java.lang.reflect.Method;

import java.net.URL;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.dom.NodeImpl;

import org.eclipse.xsd.XSDDiagnostic;
import org.eclipse.xsd.XSDDiagnosticSeverity;
import org.eclipse.xsd.XSDFactory;
import org.eclipse.xsd.XSDPackage;
import org.eclipse.xsd.XSDPlugin;
import org.eclipse.xsd.XSDSchema;

import org.eclipse.xsd.impl.XSDSchemaImpl;

import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

import org.xml.sax.helpers.DefaultHandler;

import org.xml.sax.ext.LexicalHandler;


/**
 * The <b>SAX Parser</b> for the model.
 * This handles the conversion of SAX parse exceptions to {@link XSDDiagnostic}s.
 * It also computes line column information, 
 * which is stored in a map on each node's {@link NodeImpl#getUserData user data}.
 * This class is not intended for reuse outside of the model implementation and is subject to change.
 */
public class XSDParser extends DefaultHandler implements LexicalHandler
{
  /**
   * Returns the {@link NodeImpl#getUserData user data} associated with the node.
   * If the node has no user data, a new empty map is created.
   * A <code>null</code> will be returned only in the case that 
   * the node cannot have user data 
   * or already has some other type of user data.
   * @param to the node to query.
   * @return the user data associated with the node.
   */
  public static Map getUserData(Node node)
  {
    Map result = null;
    if (node instanceof NodeImpl)
    {
      NodeImpl nodeImpl = (NodeImpl)node;
      Object userData = nodeImpl.getUserData();
      if (userData == null)
      {
        result = new HashMap();
        nodeImpl.setUserData(result);
      }
      else if (userData instanceof Map)
      {
        result = (Map)userData;
      }
    }
    return result;
  }

  /**
   * Returns the line at which the given node starts.
   * @param to the node to query.
   * @return the line at which the given node starts.
   */
  public static int getStartLine(Node node)
  {
    Integer result = (Integer)getUserData(node).get("startLine");
    return result == null ? 1 : result.intValue();
  }

  /**
   * Returns the column at which the given node starts.
   * @param to the node to query.
   * @return the column at which the given node starts.
   */
  public static int getStartColumn(Node node)
  {
    Integer result = (Integer)getUserData(node).get("startColumn");
    return result == null ? 1 : result.intValue();
  }

  /**
   * Returns the line at which the given node ends.
   * @param to the node to query.
   * @return the line at which the given node ends.
   */
  public static int getEndLine(Node node)
  {
    Integer result = (Integer)getUserData(node).get("endLine");
    return result == null ? 1 : result.intValue();
  }

  /**
   * Returns the column at which the given node ends.
   * @param to the node to query.
   * @return the column at which the given node ends.
   */
  public static int getEndColumn(Node node)
  {
    Integer result = (Integer)getUserData(node).get("endColumn");
    return result == null ? 1 : result.intValue();
  }

  protected static final String DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";


  /**
   * The cached package.
   */

  /**
   * The cached factory.
   */
  protected XSDFactory xsdFactory = XSDFactory.eINSTANCE;

  protected XSDSchema xsdSchema;
  protected List xsdDiagnostics = new ArrayList();
  protected SAXParser saxParser;
  protected Document document;
  protected Stack stack = new Stack();
  protected Element element;
  protected Locator locator;
  protected int line;
  protected int column;
  protected String encoding;

  public XSDParser()
  {
    ClassLoader previousClassLoader = Thread.currentThread().getContextClassLoader();
    try 
    {
      Thread.currentThread().setContextClassLoader(getClass().getClassLoader());
  
      // This is a gigantic hack in order to be able to access the XML encoding 
      // which is inaccessible in the current version of Xerces.
      // What we really want is simply this line of code:
      //   SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
      // For now we have this crappy thing instead.
      // BEGIN-GIANT-HACK
      //
      class GiantHack extends javax.xml.parsers.SAXParserFactory
      {
        protected java.util.Hashtable features;

        public SAXParser newSAXParser() throws ParserConfigurationException
        {
          try
          {
            return 
              new javax.xml.parsers.SAXParser()
              {
                protected XMLReader xmlReader;
                protected String schemaLanguage = null;     
                
                {
                  try
                  {
                    xmlReader = 
                      new org.apache.xerces.parsers.SAXParser()
                      {
                        class MyLocatorProxy extends LocatorProxy
                        {
                          protected String encoding;
                          MyLocatorProxy(org.apache.xerces.xni.XMLLocator locator)
                          {
                            super(locator);
                            this.encoding = encoding;
                          }

                          public String getEncoding()
                          {
                            return encoding;
                          }

                          public void setEncoding(String encoding)
                          {
                            this.encoding = encoding;
                          }
                        }

                        protected MyLocatorProxy locator;

                        public void startDocument(org.apache.xerces.xni.XMLLocator locator, String encoding, org.apache.xerces.xni.Augmentations augs) throws org.apache.xerces.xni.XNIException 
                        {
                          try 
                          {
                            if (fDocumentHandler != null) 
                            {
                              if (locator != null) 
                              {
                                fDocumentHandler.setDocumentLocator(this.locator = new MyLocatorProxy(locator));
                              }
                              fDocumentHandler.startDocument();
                            }
                         
                            if (fContentHandler != null) 
                            {
                              if (locator != null) 
                              {
                                fContentHandler.setDocumentLocator(this.locator = new MyLocatorProxy(locator));
                              }
                              fContentHandler.startDocument();
                            }
                          }
                          catch (SAXException e) 
                          {
                            throw new org.apache.xerces.xni.XNIException(e);
                          }
                        }

                        public void xmlDecl(String version, String encoding, String standalone, org.apache.xerces.xni.Augmentations augs)
                        {
                          // The encoding is not known until this point.
                          //
                          if (locator != null)
                          {
                            locator.setEncoding(encoding);
                          }

                          super.xmlDecl(version, encoding, standalone, augs);
                        }
                      };
                    xmlReader.setFeature(org.apache.xerces.impl.Constants.SAX_FEATURE_PREFIX + org.apache.xerces.impl.Constants.VALIDATION_FEATURE, GiantHack.this.isValidating());
                    xmlReader.setFeature(org.apache.xerces.impl.Constants.SAX_FEATURE_PREFIX + org.apache.xerces.impl.Constants.NAMESPACES_FEATURE, GiantHack.this.isNamespaceAware());
                    xmlReader.setFeature(org.apache.xerces.impl.Constants.SAX_FEATURE_PREFIX + org.apache.xerces.impl.Constants.NAMESPACE_PREFIXES_FEATURE, !GiantHack.this.isNamespaceAware());
                    setFeatures(features);
                  }
                  catch (Exception exception)
                  {
                    exception.printStackTrace();
                  }
                }

                private void setFeatures(java.util.Hashtable features) throws SAXNotSupportedException, SAXNotRecognizedException
                {
                  if (features != null) 
                  {
                    for (java.util.Enumeration e = features.keys(); e.hasMoreElements();) 
                    {
                      String feature = (String)e.nextElement();
                      boolean value = ((Boolean)features.get(feature)).booleanValue();
                      xmlReader.setFeature(feature, value);
                    }
                  }
                }

                public org.xml.sax.Parser getParser() throws SAXException 
                {
                  return (org.xml.sax.Parser) xmlReader;
                }

                public XMLReader getXMLReader() 
                {
                  return xmlReader;
                }

                public boolean isNamespaceAware() 
                {
                  try 
                  {
                    return xmlReader.getFeature(org.apache.xerces.impl.Constants.SAX_FEATURE_PREFIX + org.apache.xerces.impl.Constants.NAMESPACES_FEATURE);
                  } 
                  catch (SAXException x) 
                  {
                    throw new IllegalStateException(x.getMessage());
                  }
                }

                public boolean isValidating() 
                {
                  return false;
                }

                public void setProperty(String name, Object value) throws SAXNotRecognizedException, SAXNotSupportedException
                {
                  if ("http://java.sun.com/xml/jaxp/properties/schemaLanguage".equals(name)) 
                  {
                    if ("http://www.w3.org/2001/XMLSchema".equals(value)) 
                    {
                      schemaLanguage = "http://www.w3.org/2001/XMLSchema";
                      xmlReader.setFeature(org.apache.xerces.impl.Constants.XERCES_FEATURE_PREFIX + org.apache.xerces.impl.Constants.SCHEMA_VALIDATION_FEATURE, true);
                    } 
                    else if (value == null) 
                    {
                      schemaLanguage = null;
                      xmlReader.setFeature(org.apache.xerces.impl.Constants.XERCES_FEATURE_PREFIX + org.apache.xerces.impl.Constants.SCHEMA_VALIDATION_FEATURE, false);
                    } 
                    else 
                    {
                      throw new SAXNotSupportedException("Unsupported schema language");
                    }
                  } 
                  else 
                  {
                    xmlReader.setProperty(name, value);
                  }
                }

                public Object getProperty(String name) throws SAXNotRecognizedException, SAXNotSupportedException
                {
                  if ("http://java.sun.com/xml/jaxp/properties/schemaLanguage".equals(name)) 
                  {
                    return schemaLanguage;
                  } 
                  else 
                  {
                    return xmlReader.getProperty(name);
                  }
                }
              };
          }
          catch (Exception exception)
          {
            throw new ParserConfigurationException(exception.getMessage());
          }
        }

        public void setFeature(String name, boolean value)
        {
          if (features == null)
          {
            features = new java.util.Hashtable();
          }
          features.put(name, value ? Boolean.TRUE : Boolean.FALSE);
        }

        public boolean getFeature(String name)
        {
          return false;
        }
      };
      SAXParserFactory saxParserFactory = new GiantHack();
      // END-GIANT-HACK

      saxParserFactory.setNamespaceAware(true);
      saxParserFactory.setValidating(false);

      saxParserFactory.setFeature("http://xml.org/sax/features/validation", false);
      saxParserFactory.setFeature("http://xml.org/sax/features/namespaces", true);
      saxParserFactory.setFeature("http://xml.org/sax/features/namespace-prefixes", true);

      saxParser = saxParserFactory.newSAXParser();

      saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
    } 
    catch (SAXException exception)
    {
      fatalError(exception);
    }
    catch (ParserConfigurationException exception) 
    {
      fatalError(exception);
    }
    finally 
    {
      Thread.currentThread().setContextClassLoader(previousClassLoader);
    }
  }

  public void parse(String uri)
  {
    try
    {
      saxParser.parse(new InputSource(uri), this);
    }
    catch (IOException exception)
    {
      fatalError(exception);
    }
    catch (SAXException exception)
    {
      if (xsdDiagnostics.isEmpty())
      {
        fatalError(exception);
      }
    }
  }

  public void parseString(String input)
  {
    parse(new InputSource(new StringReader(input)));
  }

  public void parse(InputSource inputSource)
  {
    try
    {
      saxParser.parse(inputSource, this);
    }
    catch (IOException exception)
    {
      fatalError(exception);
    }
    catch (SAXException exception)
    {
      if (xsdDiagnostics.isEmpty())
      {
        fatalError(exception);
      }
    }
  }

  public void parse(InputStream inputStream)
  {
    try
    {
      saxParser.parse(new InputSource(inputStream), this);
    }
    catch (IOException exception)
    {
      fatalError(exception);
    }
    catch (SAXException exception)
    {
      if (xsdDiagnostics.isEmpty())
      {
        fatalError(exception);
      }
    }
  }

  public String getEncoding()
  {
    if (locator != null)
    {
      try 
      {
        Method getEncodingMethod = locator.getClass().getMethod("getEncoding", new Class[]{});
        if (getEncodingMethod != null)
        {
          encoding = (String)getEncodingMethod.invoke(locator, null);
        }
      }
      catch (Exception e) 
      {
        // If we can't find it, there's nothing we can do...
      }
    }

    return encoding;
  }

  public Document getDocument()
  {
    return document;
  }

  public void setSchema(XSDSchema xsdSchema)
  {
    if (document != null)
    {
      this.xsdSchema = xsdSchema;
      Element element = document.getDocumentElement();
      if (element != null)
      {
        xsdSchema.setElement(element);
      }
      xsdSchema.clearDiagnostics();
      XSDResourceImpl.assignDiagnostics(xsdSchema, xsdDiagnostics);
      XSDResourceImpl resource = (XSDResourceImpl)xsdSchema.eResource();
      if (resource != null && getEncoding() != null)
      {
        resource.getDefaultSaveOptions().put(XSDResourceImpl.XSD_ENCODING, getEncoding());
      }
    }
  }

  public XSDSchema getSchema()
  {
    if (xsdSchema == null && document != null)
    {
      Element element = document.getDocumentElement();
      xsdSchema = element == null ? xsdFactory.createXSDSchema() : XSDSchemaImpl.createSchema(element);
      XSDResourceImpl.assignDiagnostics(xsdSchema, xsdDiagnostics);
    }
    return xsdSchema;
  }
  
  public Collection getDiagnostics()
  {
    return xsdDiagnostics;
  }

  public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
  {
    Element newElement = document.createElementNS(uri, qName);
    for (int i = 0, length = attributes.getLength(); i < length; ++i)
    {
      String attributeURI = attributes.getURI(i);
      String attributeQName = attributes.getQName(i);
      String attributeValue = attributes.getValue(i);
      if (attributeQName.equals("xmlns") || attributeQName.startsWith("xmlns:"))
      {
        attributeURI = XSDConstants.XMLNS_URI_2000;
      }
      else if ("".equals(attributeURI))
      {
        attributeURI = null;
      }
      newElement.setAttributeNS(attributeURI, attributeQName, attributeValue);
    }

    if (stack.isEmpty())
    {
      document.appendChild(newElement);
    }
    else
    {
      element.appendChild(newElement);
    }

    stack.push(element);
    element = newElement;

    Map extendedAttributes = getUserData(element);
    extendedAttributes.put("startLine", new Integer(line));
    extendedAttributes.put("startColumn", new Integer(column));

    saveLocation();
  }

  public void endElement(String uri, String localName, String qName) throws SAXException
  {
    saveLocation();

    Map extendedAttributes = getUserData(element);
    extendedAttributes.put("endLine", new Integer(line));
    extendedAttributes.put("endColumn", new Integer(column));

    element = (Element)stack.pop();
  }

  public void setDocumentLocator(Locator locator)
  {
    this.locator = locator;
    super.setDocumentLocator(locator);
  }

  public void startDocument()
  {
    saveLocation();
    document = new DocumentImpl();
    xsdSchema = null;
  }

  public void endDocument()
  {
    element = null;
    saveLocation();
  }

  public void characters(char [] characters, int start, int length) throws SAXException
  {
    Text textNode = document.createTextNode(new String(characters, start, length));
    element.appendChild(textNode);
    saveLocation();
  }

  public void comment(char [] characters, int start, int length) throws SAXException
  {
    Comment commentNode = document.createComment(new String(characters, start, length));
    if (element == null)
    {
      document.appendChild(commentNode);
    }
    else
    {
      element.appendChild(commentNode);
    }
    saveLocation();
  }

  public void startCDATA() 
  {
  }

  public void endCDATA() 
  {
  }

  public void startDTD(java.lang.String name, java.lang.String publicId, java.lang.String systemId) 
  {
  }

  public void endDTD() 
  {
  }

  public void startEntity(java.lang.String name) 
  {
  }

  public void endEntity(java.lang.String name) 
  {
  }

  protected void fatalError(IOException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.FATAL_LITERAL);
    xsdDiagnostic.setMessage(XSDPlugin.INSTANCE.getString("_UI_IOError_message", new Object [] { exception.getLocalizedMessage() }));
    xsdDiagnostics.add(xsdDiagnostic);
  }

  protected void fatalError(ParserConfigurationException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.FATAL_LITERAL);
    xsdDiagnostic.setMessage(XSDPlugin.INSTANCE.getString("_UI_ParserError_message", new Object [] { exception.getLocalizedMessage() }));
    xsdDiagnostics.add(xsdDiagnostic);
  }

  public void fatalError(SAXException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.FATAL_LITERAL);
    xsdDiagnostic.setMessage(XSDPlugin.INSTANCE.getString("_UI_ParserError_message", new Object [] { exception.getMessage() }));
    xsdDiagnostics.add(xsdDiagnostic);
  }

  public void fatalError(SAXParseException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.FATAL_LITERAL);
    xsdDiagnostic.setMessage(XSDPlugin.INSTANCE.getString("_UI_ParserError_message", new Object [] { exception.getMessage() }));
    xsdDiagnostic.setLine(exception.getLineNumber());
    xsdDiagnostic.setColumn(exception.getColumnNumber());
    xsdDiagnostics.add(xsdDiagnostic);
  }

  public void error(SAXParseException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.ERROR_LITERAL);
    xsdDiagnostic.setMessage("DOM:" + exception.getMessage());
    xsdDiagnostic.setLine(exception.getLineNumber());
    xsdDiagnostic.setColumn(exception.getColumnNumber());
    xsdDiagnostics.add(xsdDiagnostic);
  }

  public void warning(SAXParseException exception)
  {
    XSDDiagnostic xsdDiagnostic = xsdFactory.createXSDDiagnostic();
    xsdDiagnostic.setSeverity(XSDDiagnosticSeverity.WARNING_LITERAL);
    xsdDiagnostic.setMessage("DOM:" + exception.getMessage());
    xsdDiagnostic.setLine(exception.getLineNumber());
    xsdDiagnostic.setColumn(exception.getColumnNumber());
    xsdDiagnostics.add(xsdDiagnostic);
  }

  protected void saveLocation()
  {
    line = locator.getLineNumber();
    column = locator.getColumnNumber();
  }

  public InputSource resolveEntity(String publicId, String systemId) throws SAXException
  {
    InputSource inputSource;
    if ("-//W3C//DTD XMLSCHEMA 200102//EN".equals(publicId))
    {
      inputSource = new InputSource( XSDPlugin.INSTANCE.getBaseURL() + "cache/www.w3.org/2001/XMLSchema.dtd");
      inputSource.setPublicId(publicId);
    }
    else
    {
      inputSource = super.resolveEntity(publicId, systemId);
    }

    return inputSource;
  }
}
